{ SysTray component version 1.1 copyright (c) 2000 Michal Cihar

LICENSE:
  This component may be used without any charge in non-commercial
applications. For commercial applications you must contact author
and recieve premission to use this. Complete license text can be
found in file LICENSE.TXT which should you get with this component.

CONTACTING AUTHOR:
  Email: cihar@email.cz
  WWW: http://miso.tsx.org
  ICQ: 33471395

DESCRIPTION:
  SysTray component can handle following things:
    - displays icon in system tray, adds popup menus and events
      to this icon
    - hides taskbar button for application
    - can minimize applicaion to system tray

PROPERTIES:
  Published; read/write:
    Icon - icon displayed in system tray (if Visible true)
    Hint - hint for system tray icon, when changed (ONLY) in design state
          automatically adjusts ShowHint to true (anything entered) or false
    ShowHint - if true hint is diplayed
    PopupLeft, PopupRight - popup menus for each button, to display must be
                            AutoPopup property (of PopupMenu) set to true
    Visible - if true icon is displayed
    StartMinimized - starts application without displaying main form
    TaskHide - if true application is hidded from taskbar
    MinimizeToTray - if true minimalising application causes hiding it to tray
  Warning: when you set Visible to false and StartMinimized to true you haven't
           almost ANY chance to control your program...

EVENTS:
     OnClick, OnDblClick, OnMouseDown, OnMouseMove, OnMouseUp, OnLeftClick,
     OnLeftDblClick, OnMiddleClick, OnMiddleDblClick, OnRightClick, OnRightDblClick
      - here is nothing to describe - everything clear...

OVERRIDES:
     DoMinimizeToTray - change this to customize behavior of application while
                        minimizing to tray

CHANGES:
    Version 1.1 (14.9.2000)
                * fixed some small bugs in code
                * some code optimalisations
                + added hook to application and form to catch minimize actions
                + added MinimizeToTray property
                + added DoMinimizeToTray procedure, which cen be oerridden to
                  customize behavior of application while minimizing to tray
                + added events to distinguish which button was pressed
                * popup menu is displayed only if it has AutoPopup property
                  set to true
                * when changing Hint (ONLY) in design state, ShowHint
                  automatically adjusts to true (anything entered) or false
                  (Hint contains only spaces or is empty)
    Version 1.0 (29.1.2000)
                * first release (almost not public)


ACKNOWLEDGEMENT:
    I can't hide that this component uses some small parts from other simmilar
    componets (such as CoolTrayIcon by Troels Jakobsen - tjak@get2net.dk).
    I made this because of any of simmilar components I found didn't fit
    exactly my wishes. 
}
unit SysTray;

interface

uses
   Classes, Windows, Messages, Graphics, Menus, Controls, ShellAPI, Forms;

type
   TSysTray = class(TComponent)
   private
      FHandle: HWND;
      FShift: TShiftState;
      FWindowStyle: longint;
      FDown: boolean;

      FIcon: TIcon;
      FPopupLeft: TPopupMenu;
      FPopupRight: TPopupMenu;

      FHint: string;
      FShowHint: boolean;
      FTaskHide: boolean;
      FVisible: boolean;
      FStartMinimized: boolean;
      FMinimizeToTray: boolean;

      FOnClick: TNotifyEvent;
      FOnDblClick: TNotifyEvent;

      FOnLeftClick: TNotifyEvent;
      FOnLeftDblClick: TNotifyEvent;
      FOnMiddleClick: TNotifyEvent;
      FOnMiddleDblClick: TNotifyEvent;
      FOnRightClick: TNotifyEvent;
      FOnRightDblClick: TNotifyEvent;

      FOnMouseDown: TMouseEvent;
      FOnMouseUp: TMouseEvent;
      FOnMouseMove: TMouseMoveEvent;

      OldAppProc, NewAppProc: Pointer; // Procedure variables
      OldWndProc, NewWndProc: Pointer; // Procedure variables

      procedure WndProc(var AMessage: TMessage);

      procedure HookApp;
      procedure UnhookApp;
      procedure HookAppProc(var Msg: TMessage);

      procedure HookParent;
      procedure UnhookParent;
      procedure HookWndProc(var Msg: TMessage);

      procedure HideApplication;

      procedure SetHint(AValue: string);
      procedure SetIcon(AValue: TIcon);
      procedure SetShowHint(AValue: boolean);
      procedure ShowIcon(AValue: boolean);
      procedure SetTaskHide(AValue: boolean);
   protected
      IconData: TNotifyIconData;
   public
      constructor Create(AOwner: TComponent); override;
      destructor Destroy; override;
      procedure Loaded; override;
      procedure DoMinimizeToTray; dynamic;
   published
      property Hint: string read FHint write SetHint;
      property Icon: TIcon read FIcon write SetIcon;
      property PopupLeft: TPopupMenu read FPopupLeft write FPopupLeft;
      property PopupRight: TPopupMenu read FPopupRight write FPopupRight;
      property ShowHint: boolean read FShowHint write SetShowHint default false;
      property Visible: boolean read FVisible write ShowIcon default true;
      property StartMinimized: boolean read FStartMinimized write FStartMinimized default false;
      property TaskHide: boolean read FTaskHide write SetTaskHide default false;
      property MinimizeToTray: boolean read FMinimizeToTray write FMinimizeToTray default true;

      property OnClick: TNotifyEvent read FOnClick write FOnClick default nil;
      property OnDblClick: TNotifyEvent read FOnDblClick write FOnDblClick default nil;

      property OnLeftClick: TNotifyEvent read FOnLeftClick write FOnLeftClick default nil;
      property OnLeftDblClick: TNotifyEvent read FOnLeftDblClick write FOnLeftDblClick default nil;
      property OnMiddleClick: TNotifyEvent read FOnMiddleClick write FOnMiddleClick default nil;
      property OnMiddleDblClick: TNotifyEvent read FOnMiddleDblClick write FOnMiddleDblClick default
         nil;
      property OnRightClick: TNotifyEvent read FOnRightClick write FOnRightClick default nil;
      property OnRightDblClick: TNotifyEvent read FOnRightDblClick write FOnRightDblClick default
         nil;

      property OnMouseDown: TMouseEvent read FOnMouseDown write FOnMouseDown default nil;
      property OnMouseMove: TMouseMoveEvent read FOnMouseMove write FOnMouseMove default nil;
      property OnMouseUp: TMouseEvent read FOnMouseUp write FOnMouseUp default nil;
   end;

procedure Register;

implementation

uses
   SysUtils;

var
   WM_SysTrayIcon: UINT;

constructor TSysTray.Create;
begin
   inherited Create(AOwner);
   FDown := False;

   FIcon := TIcon.Create;

   FShowHint := false;
   FVisible := true;
   FStartMinimized := false;
   FTaskHide := false;
   FMinimizeToTray := true;

   if not (csDesigning in ComponentState) then
   begin
      try
         FHandle := AllocateHWND(WndProc);

         IconData.cbSize := SizeOf(IconData);
         IconData.uID := 1;
         IconData.uCallbackMessage := WM_SysTrayIcon;
         IconData.Wnd := FHandle;

         IconData.uFlags := NIF_MESSAGE;
         if Assigned(FIcon) then
         begin
            IconData.uFlags := (IconData.uFlags or NIF_ICON);
            IconData.hIcon := FIcon.Handle;
         end;
         if FShowHint then
         begin
            IconData.uFlags := (IconData.uFlags or NIF_TIP);
            StrCopy(IconData.szTip, PChar(FHint));
         end;
      finally
         Shell_NotifyIcon(NIM_ADD, @IconData)
      end;
      HookApp;
      HookParent;
   end;
end;

destructor TSysTray.Destroy;
begin
   if not (csDesigning in ComponentState) then
   begin
      UnhookApp;
      UnhookParent;
      DeallocateHWND(FHandle);
      Shell_NotifyIcon(NIM_DELETE, @IconData);
   end;
   FIcon.Free;
   inherited Destroy;
end;

procedure TSysTray.Loaded;
begin
   inherited Loaded;
   if not (csDesigning in ComponentState) then
   begin
      IconData.uFlags := NIF_MESSAGE;
      if Assigned(FIcon) then
      begin
         IconData.uFlags := (IconData.uFlags or NIF_ICON);
         IconData.hIcon := FIcon.Handle;
      end;
      if FShowHint then
      begin
         IconData.uFlags := (IconData.uFlags or NIF_TIP);
         StrCopy(IconData.szTip, PChar(FHint));
      end;

      Shell_NotifyIcon(NIM_MODIFY, @IconData);

      if FStartMinimized then
      begin
         Application.ShowMainForm := False;
         ShowWindow(Application.Handle, SW_HIDE);
      end;
   end;
end;

procedure TSysTray.HookApp;
begin
   OldAppProc := Pointer(GetWindowLong(Application.Handle, GWL_WNDPROC));
   NewAppProc := MakeObjectInstance(HookAppProc);
   SetWindowLong(Application.Handle, GWL_WNDPROC, LongInt(NewAppProc));
end;


procedure TSysTray.UnhookApp;
begin
   if Assigned(OldAppProc) then
      SetWindowLong(Application.Handle, GWL_WNDPROC, LongInt(OldAppProc));
   if Assigned(NewAppProc) then
      FreeObjectInstance(NewAppProc);
   NewAppProc := nil;
   OldAppProc := nil;
end;


procedure TSysTray.HookAppProc(var Msg: TMessage);
begin
   case Msg.Msg of
      WM_SIZE:
         if Msg.wParam = SIZE_MINIMIZED then
         begin
            if FMinimizeToTray then
               DoMinimizeToTray;
         end;
   end;
   Msg.Result := CallWindowProc(OldAppProc, Application.Handle, Msg.Msg, Msg.wParam, Msg.lParam);
end;

procedure TSysTray.HookParent;
begin
   if (Owner as TWinControl) <> nil then
   begin
      OldWndProc := Pointer(GetWindowLong((Owner as TWinControl).Handle, GWL_WNDPROC));
      NewWndProc := MakeObjectInstance(HookWndProc);
      SetWindowLong((Owner as TWinControl).Handle, GWL_WNDPROC, LongInt(NewWndProc));
   end;
end;


procedure TSysTray.UnhookParent;
begin
   if ((Owner as TWinControl) <> nil) and Assigned(OldWndProc) then
      SetWindowLong((Owner as TWinControl).Handle, GWL_WNDPROC, LongInt(OldWndProc));
   if Assigned(NewWndProc) then
      FreeObjectInstance(NewWndProc);
   NewWndProc := nil;
   OldWndProc := nil;
end;


procedure TSysTray.HookWndProc(var Msg: TMessage);
begin
   case Msg.Msg of
      WM_SHOWWINDOW:
         begin
            if (Msg.lParam = 0) and (Msg.wParam = 1) then
            begin
               ShowWindow(Application.Handle, SW_RESTORE);
               SetForegroundWindow(Application.Handle);
               SetForegroundWindow((Owner as TWinControl).Handle);
            end;
         end;
      WM_WINDOWPOSCHANGED:
         begin
            if Assigned(Screen.ActiveControl) then
               SetFocus(Screen.ActiveControl.Handle);
         end;
   end;
   Msg.Result := CallWindowProc(OldWndProc, (Owner as TWinControl).Handle,
      Msg.Msg, Msg.wParam, Msg.lParam);
end;

procedure TSysTray.DoMinimizeToTray;
begin
   HideApplication;
end;

procedure TSysTray.HideApplication;
begin
   ShowWindow(Application.Handle, SW_HIDE);
   Application.MainForm.Hide;
   if (Owner is TWinControl) then
      (Owner as TWinControl).Hide;
end;


procedure TSysTray.WndProc;
var
   CursorPos: TPoint;

   function ShiftState: TShiftState;
 {shift state ONLY for keys}
   begin
      Result := [];
      if GetKeyState(VK_SHIFT) < 0 then Include(Result, ssShift);
      if GetKeyState(VK_CONTROL) < 0 then Include(Result, ssCtrl);
      if GetKeyState(VK_MENU) < 0 then Include(Result, ssAlt);
   end;

begin
   if AMessage.Msg = WM_SysTrayIcon then
   begin
      GetCursorPos(CursorPos);
      case AMessage.lParam of
         WM_RBUTTONDBLCLK:
            begin
               FDown := False;
               FShift := ShiftState;
               if Assigned(FOnDblClick) then FOnDblClick(Self);
               if Assigned(FOnRightDblClick) then FOnRightDblClick(Self);
            end;
         WM_MBUTTONDBLCLK:
            begin
               FDown := False;
               FShift := ShiftState;
               if Assigned(FOnDblClick) then FOnDblClick(Self);
               if Assigned(FOnMiddleDblClick) then FOnMiddleDblClick(Self);
            end;
         WM_LBUTTONDBLCLK:
            begin
               FDown := False;
               FShift := ShiftState;
               if Assigned(FOnDblClick) then FOnDblClick(Self);
               if Assigned(FOnLeftDblClick) then FOnLeftDblClick(Self);
            end;
         WM_LBUTTONDOWN:
            begin
               FDown := True;
               FShift := ShiftState + [ssLeft];
               if Assigned(FPopupLeft) then
                  if FPopupLeft.AutoPopup then
                  begin
                     if (Owner is TWinControl) then
                        SetForegroundWindow((Owner as TWinControl).Handle);
                     FPopupLeft.PopupComponent := Self;
                     FPopupLeft.Popup(CursorPos.X, CursorPos.Y);
                     PostMessage((Owner as TWinControl).Handle, WM_NULL, 0, 0);
                  end;
               if Assigned(FOnMouseDown) then
                  FOnMouseDown(Self, mbLeft, FShift, CursorPos.X, CursorPos.Y);
            end;
         WM_RBUTTONDOWN:
            begin
               FDown := True;
               FShift := ShiftState + [ssRight];
               if Assigned(FPopupRight) then
                  if FPopupRight.AutoPopup then
                  begin
                     if (Owner is TWinControl) then
                        SetForegroundWindow((Owner as TWinControl).Handle);
                     FPopupRight.PopupComponent := Self;
                     FPopupRight.Popup(CursorPos.X, CursorPos.Y);
                     PostMessage((Owner as TWinControl).Handle, WM_NULL, 0, 0);
                  end;
               if Assigned(FOnMouseDown) then
                  FOnMouseDown(Self, mbRight, FShift, CursorPos.X, CursorPos.Y);
            end;
         WM_MBUTTONDOWN:
            begin
               FDown := True;
               FShift := ShiftState + [ssMiddle];
               if Assigned(FOnMouseDown) then
                  FOnMouseDown(Self, mbMiddle, FShift, CursorPos.X, CursorPos.Y);
            end;
         WM_LBUTTONUP:
            begin
               FShift := ShiftState;
               if Assigned(FOnMouseUp) then
                  FOnMouseUp(Self, mbLeft, FShift, CursorPos.X, CursorPos.Y);
               if FDown and Assigned(FOnClick) then FOnClick(Self);
               if FDown and Assigned(FOnLeftClick) then FOnLeftClick(Self);
               FDown := False;
            end;
         WM_RBUTTONUP:
            begin
               FShift := ShiftState;
               if Assigned(FOnMouseUp) then
                  FOnMouseUp(Self, mbRight, FShift, CursorPos.X, CursorPos.Y);
               if FDown and Assigned(FOnClick) then FOnClick(Self);
               if FDown and Assigned(FOnRightClick) then FOnRightClick(Self);
               FDown := False;
            end;
         WM_MBUTTONUP:
            begin
               FShift := ShiftState;
               if Assigned(FOnMouseUp) then
                  FOnMouseUp(Self, mbMiddle, FShift, CursorPos.X, CursorPos.Y);
               if FDown and Assigned(FOnClick) then FOnClick(Self);
               if FDown and Assigned(FOnMiddleClick) then FOnMiddleClick(Self);
               FDown := False;
            end;

         WM_MOUSEMOVE:
            begin
               if GetKeyState(VK_SHIFT) < 0 then Include(FShift, ssShift);
               if GetKeyState(VK_CONTROL) < 0 then Include(FShift, ssCtrl);
               if GetKeyState(VK_MENU) < 0 then Include(FShift, ssAlt);
               if Assigned(FOnMouseMove) then FOnMouseMove(Self, FShift, CursorPos.X, CursorPos.Y);
            end;
      end;
   end
   else
      AMessage.Result := DefWindowProc(FHandle, AMessage.Msg, AMessage.WParam, AMessage.LParam);
end;



procedure TSysTray.SetIcon;
begin
   FIcon.Assign(AValue);
   if (not (csDesigning in ComponentState)) and (not (FIcon.Empty)) then
   begin
      IconData.uFlags := (IconData.uFlags or NIF_ICON);
      IconData.hIcon := FIcon.Handle;
      Shell_NotifyIcon(NIM_MODIFY, @IconData);
   end
   else
   begin
      IconData.uFlags := IconData.uFlags and not (NIF_ICON);
      Shell_NotifyIcon(NIM_MODIFY, @IconData);
   end
end;

procedure TSysTray.SetHint;
begin
   FHint := AValue;
   if (not (csDesigning in ComponentState)) then
   begin
      if (FShowHint) then
      begin
         IconData.uFlags := (IconData.uFlags or NIF_TIP);
         StrCopy(IconData.szTip, PChar(FHint));
         Shell_NotifyIcon(NIM_MODIFY, @IconData);
      end
      else
      begin
         IconData.uFlags := IconData.uFlags and not (NIF_TIP);
         Shell_NotifyIcon(NIM_MODIFY, @IconData);
      end;
   end
   else
      ShowHint := trim(FHint) <> '';
end;

procedure TSysTray.SetShowHint;
begin
   FShowHint := AValue;
   if (not (csDesigning in ComponentState)) and (FShowHint) then
   begin
      IconData.uFlags := (IconData.uFlags or NIF_TIP);
      StrCopy(IconData.szTip, PChar(FHint));
      Shell_NotifyIcon(NIM_MODIFY, @IconData);
   end
   else
   begin
      IconData.uFlags := IconData.uFlags and not (NIF_TIP);
      Shell_NotifyIcon(NIM_MODIFY, @IconData);
   end
end;

procedure TSysTray.SetTaskHide;
begin
   if FTaskHide <> AValue then
   begin
      FTaskHide := AValue;
      if not (csDesigning in ComponentState) then
      begin
         ShowWindow(Application.Handle, SW_HIDE);
         if FTaskHide then
         begin
            FWindowStyle := GetWindowLong(Application.Handle, GWL_EXSTYLE);
            SetWindowLong(Application.Handle, GWL_EXSTYLE, FWindowStyle or WS_EX_TOOLWINDOW and not
               WS_EX_APPWINDOW or WS_EX_TOPMOST)
         end
         else
            SetWindowLong(Application.Handle, GWL_EXSTYLE, FWindowStyle);
         ShowWindow(Application.Handle, SW_SHOW);
      end;
   end;
end;

procedure TSysTray.ShowIcon;
begin
   if not (csDesigning in ComponentState) then
   begin
      if (FVisible <> AValue) then
      begin
         if AValue then
            Shell_NotifyIcon(NIM_ADD, @IconData)
         else
            Shell_NotifyIcon(NIM_DELETE, @IconData);
         FVisible := AValue;
      end
      else
         if FVisible then Shell_NotifyIcon(NIM_MODIFY, @IconData);
   end;
end;

procedure Register;
begin
   RegisterComponents('MiSo', [TSysTray]);
end;

initialization
   begin
      WM_SysTrayIcon := RegisterWindowMessage('SysTrayIcon');
   end;

end.

